#include "slippyMap.h"
#include "constants.h"

#include <QNetworkDiskCache>
#include <QDesktopServices>

#include "log.h"


SlippyMap::SlippyMap(QNetworkSession *session, QObject *parent = 0, Log* log=0, int zoom=0, MapType mapType=MapTypeOpenStreetMaps) :
    QObject(parent),
    width(400),
    height(300),
    zoom(zoom),
    latitude(0),
    longitude(0),
    network(false),
    m_session(session),
    log(log),mapType(mapType){

    m_emptyTile = QPixmap(tdim, tdim);
    m_emptyTile.fill(Qt::lightGray);

    m_manager = new QNetworkAccessManager(this);

    //log->debug(QString("Is Network accessible?:%1").arg(m_manager->networkAccessible()));
    QNetworkDiskCache *cache = new QNetworkDiskCache;
    cache->setCacheDirectory(QDesktopServices::storageLocation(QDesktopServices::CacheLocation));
    m_manager->setCache(cache);
    connect(m_manager, SIGNAL(finished(QNetworkReply*)),
            this, SLOT(handleNetworkData(QNetworkReply*)));

    //log->debug("exiting from slippyMap constructor!!!!");

}

SlippyMap::~SlippyMap() {
    for (int i = 0; i < m_pendingReplies.size(); ++i) {
        delete m_pendingReplies.at(i);
    }
}


void SlippyMap::invalidate(qreal x, qreal y){


    double tx = x;
    double ty = y;

    // top-left corner of the center tile

    int xp = width / 2 - (tx - floor(tx)) * tdim;
    int yp = height / 2 - (ty - floor(ty)) * tdim;
    //log->info(QString("xp:%1=%2/2-(%3 - %4)*%5").arg(xp).arg(width).arg(tx).arg(floor(tx)).arg(tdim));
    //log->info(QString("yp:%1=%2/2-(%3 - %4)*%5").arg(yp).arg(height).arg(ty).arg(floor(ty)).arg(tdim));


    // first tile vertical and horizontal
    int xa = (xp + tdim - 1) / tdim;
    int ya = (yp + tdim - 1) / tdim;
    int xs = static_cast<int>(tx) - xa;
    int ys = static_cast<int>(ty) - ya;

    //log->info(QString("xa=%1").arg(xa));
    //log->info(QString("ya=%1").arg(ya));

    //log->info(QString("xs=%1").arg(xs));
    //log->info(QString("ys=%1").arg(ys));


    // offset for top-left tile
    m_offset = QPoint(xp - xa * tdim, yp - ya * tdim);

    // last tile vertical and horizontal
    int xe = static_cast<int>(tx) + (width - xp - 1) / tdim;
    int ye = static_cast<int>(ty) + (height - yp - 1) / tdim;

    // build a rect
    m_tilesRect = QRect(xs, ys, xe - xs + 1, ye - ys + 1);

    if (m_url.isEmpty())
        download();

    emit updated(QRect(0, 0, width, height));
}

void SlippyMap::invalidate(){
    //log->debug("INVALIDATE");
    if (width <= 0 || height <= 0 || (latitude==0 && longitude==0))
        return;

    QPointF ct = tileForCoordinate(latitude, longitude, zoom);
    double tx = ct.x();
    double ty = ct.y();

    // top-left corner of the center tile

    int xp = width / 2 - (tx - floor(tx)) * tdim;
    int yp = height / 2 - (ty - floor(ty)) * tdim;
    //log->info(QString("xp:%1=%2/2-(%3 - %4)*%5").arg(xp).arg(width).arg(tx).arg(floor(tx)).arg(tdim));
    //log->info(QString("yp:%1=%2/2-(%3 - %4)*%5").arg(yp).arg(height).arg(ty).arg(floor(ty)).arg(tdim));


    // first tile vertical and horizontal
    int xa = (xp + tdim - 1) / tdim;
    int ya = (yp + tdim - 1) / tdim;
    int xs = static_cast<int>(tx) - xa;
    int ys = static_cast<int>(ty) - ya;

    //log->info(QString("xa=%1").arg(xa));
    //log->info(QString("ya=%1").arg(ya));

    //log->info(QString("xs=%1").arg(xs));
    //log->info(QString("ys=%1").arg(ys));


    // offset for top-left tile
    m_offset = QPoint(xp - xa * tdim, yp - ya * tdim);

    // last tile vertical and horizontal
    int xe = static_cast<int>(tx) + (width - xp - 1) / tdim;
    int ye = static_cast<int>(ty) + (height - yp - 1) / tdim;

    // build a rect
    m_tilesRect = QRect(xs, ys, xe - xs + 1, ye - ys + 1);

    if (m_url.isEmpty())
        download();

    emit updated(QRect(0, 0, width, height));
}


void SlippyMap::render(QPainter *p, const QRect &rect) {
    //log->debug("rendering maps...");
    for (int x = 0; x <= m_tilesRect.width(); ++x)
        for (int y = 0; y <= m_tilesRect.height(); ++y) {
            QPoint tp(x + m_tilesRect.left(), y + m_tilesRect.top());
            QRect box = tileRect(tp);
            if (rect.intersects(box)) {
                if (m_tilePixmaps.contains(tp))
                    p->drawPixmap(box, m_tilePixmaps.value(tp));
                else
                    p->drawPixmap(box, m_emptyTile);
            }
        }
    //log->debug("done");
}

void SlippyMap::pan(const QPoint &delta) {
    QPointF dx = QPointF(delta) / qreal(tdim);
    QPointF center = tileForCoordinate(latitude, longitude, zoom) - dx;
    latitude = latitudeFromTile(center.y(), zoom);
    longitude = longitudeFromTile(center.x(), zoom);
    invalidate();
}

void SlippyMap::updatePosition(GpsPoint point) {
    latitude = point.getLatitude();
    longitude = point.getLongitude();
    invalidate();
}

void SlippyMap::clearAllMaps(){

    m_pendingReplies.clear();
    m_tilePixmaps.clear();
    invalidate();
}

void SlippyMap::handleNetworkData(QNetworkReply *reply) {
    //log->info("----- handleNetworkData -----");
    QImage img;
    QPoint tp = reply->request().attribute(QNetworkRequest::User).toPoint();
    QUrl url = reply->url();
    if (!reply->error())
        if (!img.load(reply, 0))
            img = QImage();

    m_pendingReplies.removeAll(reply);

    reply->deleteLater();
    m_tilePixmaps[tp] = QPixmap::fromImage(img);
    if (img.isNull())
        m_tilePixmaps[tp] = m_emptyTile;
    emit updated(tileRect(tp));

    // purge unused spaces
    QRect bound = m_tilesRect.adjusted(-2, -2, 2, 2);
    foreach(QPoint tp, m_tilePixmaps.keys())
    if (!bound.contains(tp))
        m_tilePixmaps.remove(tp);

    download();
}


void SlippyMap::download() {
    //log->debug("--- download function ---");
    QPoint grab(0, 0);
    for (int x = 0; x <= m_tilesRect.width(); ++x)
        for (int y = 0; y <= m_tilesRect.height(); ++y) {
            QPoint tp = m_tilesRect.topLeft() + QPoint(x, y);
            if (!m_tilePixmaps.contains(tp)) {
                grab = tp;
                break;
            }
        }
    if (grab == QPoint(0, 0)) {
        m_url = QUrl();
        return;
    }

    //log->debug(QString("network:%1").arg(network));
    if(network){

        //QString path ="http://norma.icc.cat/tilecache/tilecache.py/1.0.0/topo3857/%1/%2/%3.png?type=google";
        //QString path = "http://tile.opencyclemap.org/cycle/%1/%2/%3.png";
        //QString path = "http://tile.openstreetmap.org/%1/%2/%3.png";
        QString path = mapUrlProvider();
        m_url = QUrl(path.arg(zoom).arg(grab.x()).arg(grab.y()));
        QNetworkRequest request;

        log->info(QString("making a url request:").append(m_url.toString()));

        request.setUrl(m_url);
        request.setRawHeader("User-Agent", "GPSSniffer 1.0");
        request.setAttribute(QNetworkRequest::User, QVariant(grab));
        m_pendingReplies << m_manager->get(request);
        //log->debug("request done.");
    }else{
        log->warn("Session not opened, no making url request");
    }
}

QRect SlippyMap::tileRect(const QPoint &tp) {
    QPoint t = tp - m_tilesRect.topLeft();
    int x = t.x() * tdim + m_offset.x();
    int y = t.y() * tdim + m_offset.y();

    return QRect(x, y, tdim, tdim);
}

QString SlippyMap::mapUrlProvider(){

    //log->debug("--- mapUrlProvider ---");
    QString url("http://");
    //log->debug(QString("mapType:%1").arg(mapType));
    switch(mapType){
    case MapTypeGoogleMaps:
        url.append("mt1.google.com/vt/&x=%2&y=%3&z=%1");
        break;
    case MapTypeICC:
        url.append("norma.icc.cat/tilecache/tilecache.py/1.0.0/topo3857/%1/%2/%3.png?type=google");
        break;
    case MapTypeOpenCycleMaps:
        url.append("tile.opencyclemap.org/cycle/%1/%2/%3.png");
        break;
    default:
    case MapTypeOpenStreetMaps:
        url.append("tile.openstreetmap.org/%1/%2/%3.png");
        break;
    }
    return url;
}


